#!/usr/bin/env python

import time
import sys
from operator import itemgetter

def fileLen(fname):
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1

def init(buffer):
   from glob import globs, setGlobals
   from array import array

   # create spatial variables of specified number
   temp = buffer.pop(0)
   Vars = int(temp.split()[0].strip())+2
   TypeId = temp.split()[1].strip()
   
   # initialize constraints
   N = fileLen(globs['calc']+"/"+globs['calc']+".relations")
   setGlobals("size", N)
   Id = 0
   from bitcoding import B_dict
   with open(globs['calc']+"/"+globs['calc']+".identity") as f:
      Id = B_dict[f.readline().strip()]
  
   ConMatrix = [None for v in xrange(Vars)]
   for v in xrange(Vars):
      j = {}   
      j[v] = Id
      ConMatrix[v] = j

   # parse spatial CSP
   from parsecsp import parsecsp
   parsecsp(ConMatrix, buffer)
   
   return TypeId, ConMatrix

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg

# main function
def main(argv=None):

   if argv is None:
        argv = sys.argv

   # set some sys useful stuff
   sys.setrecursionlimit(1<<30)
   sys.setcheckinterval(10000)

   from optparse import OptionParser

   # initialize parser for command line arguments
   parser = OptionParser()
   
   # set parsing options
   parser.add_option("-c", "--calculus", type="string", dest="calc", default="rcc8", help="interaction mode: rcc8, or allen [default: %default]")
   parser.add_option("-p", "--pcmethod", type="string", dest="pcmethod", default="full", help="interaction mode: full, or incremental [default: %default]")
   parser.add_option("-d", "--print", action="store_true", dest="printsol", help="print solution to output")
   parser.add_option("-f", "--file", action="store", type="string", dest="filename")

   # parse command line arguments
   options, args = parser.parse_args()

   from glob import setGlobals
   #import argparse

   try:   
      # check and set globals for calculus information
      if not (options.calc == "rcc8" or options.calc == "allen"):
         raise Usage("Invalid option: " + options.calc)
      else:
         setGlobals('calc',options.calc)

      # check and set globals for patch consistency heuristic information
      if options.pcmethod == "full":
         setGlobals('pcmethod',0)
      elif options.pcmethod == "incremental":
         setGlobals('pcmethod',1)
      else:
         raise Usage("Invalid option: " + options.pcheuristic)

   except Usage, err:
        print err.msg
        parser.print_help()
        return 1

   with open(options.filename) as f:
      total = 0
      totalc = 0
      totali = 0
      meter = 0
      meteri = 0
      edges = 0
      edgesc = 0
      edgesi = 0
      cons = 0
      consc = 0
      consi = 0
      buffer = []

      for i in f:
         buffer.append(i)

         if i.strip() == '.':
            start = elapsed = 0
            meter += 1

            from counters import counter

            TypeId, ConMatrix = init(buffer)
            buffer = []

            from triangulation import MCS, FIC
            from bitcoding import B_dict
               
            edjes = set([])

            neighbors = tuple([set([]) for i in xrange(len(ConMatrix))])
            for n, i in enumerate(ConMatrix):
               for j in i:
                  if j > n:
                     edjes.add((n,j))
                     neighbors[n].add(j)
                     neighbors[j].add(n)

            from collections import deque
   
            fill = FIC(ConMatrix, neighbors, MCS(ConMatrix, neighbors)) ^ edjes
            totalEdjes = fill | edjes
            feeding = deque(sorted(edjes,key=itemgetter(1)))
            filling = deque(sorted(fill,key=itemgetter(1)))
            nodes = xrange(len(ConMatrix))

            from bitcoding import B_dict
            DALL = B_dict['DALL']
            for i,j in fill:
               neighbors[i].add(j)
               neighbors[j].add(i)

            for i,j in fill:
               ConMatrix[i][j] = DALL
               ConMatrix[j][i] = DALL
                  
            from glob import globs
            if globs['pcmethod'] == 0: # path consistency
               from pc import PC as pc

               start = time.clock()
               solution = pc(ConMatrix,neighbors)
               elapsed = time.clock() - start
               if solution:
                  totalc += elapsed
               else:
                  totali += elapsed
                  meteri += 1 
               total += elapsed

            elif globs['pcmethod'] == 1: # incremental path consistency
               from ipc import iPC as pc

               with open(globs['calc']+"/"+globs['calc']+".identity") as f:
                  Id = B_dict[f.readline().strip()]
 
               
               iConMatrix = []
               ineighbors = []

               elapsed = 0

               for j in nodes:

                  elapsed2 = 0

                  length = len(iConMatrix)

                  for e in xrange(length,j+1):
                     iConMatrix.append({e:Id})
                     ineighbors.append(set([]))

                  fills = []
                  while filling:
                     n,m = filling.popleft()
                     if m <= j:
                        fills.append((n,m))
                     else:
                        filling.appendleft((n,m))
                        break                    
                     
                  for f,g in fills:
                     iConMatrix[f][g] = DALL
                     iConMatrix[g][f] = DALL
                     ineighbors[f].add(g)
                     ineighbors[g].add(f)  

                  feeds = []
                  while feeding:
                     n,m = feeding.popleft()
                     if m <= j:
                        feeds.append((n,m))
                     else:
                        feeding.appendleft((n,m))
                        break                

                  for f,g in feeds:
                     iConMatrix[f][g] = ConMatrix[f][g]
                     iConMatrix[g][f] = ConMatrix[g][f]
                     ineighbors[f].add(g)
                     ineighbors[g].add(f)

                  start = time.clock()
                  solution = pc(iConMatrix,ineighbors,feeds)
                  elapsed2 += time.clock() - start
                  elapsed += elapsed2
                  total += elapsed2

                  if not solution: 
                     totali += elapsed
                     meteri += 1
                     break
               else:
                  totalc += elapsed


            # print values of complexity counters
            print '--------------------------------------'
            print 'Type ID: ', TypeId
            print '--------------------------------------'
            print 'Visited arcs: ', counter['arcs']
            print 'Checked constraints: ', counter['cons']
            print '--------------------------------------'
            print 'Time: ', elapsed, 'sec'
            print '--------------------------------------'

            # print solution  
            if solution != False:
               edgesc += counter['arcs']
               consc += counter['cons']
               print 'Path Consistent'
               if options.printsol == True:
                  from printm import printMatrix
                  if globs['pcmethod'] == 0:
                     printMatrix(ConMatrix)
                  elif globs['pcmethod'] == 1:
                     printMatrix(iConMatrix)
            else:
               edgesi += counter['arcs']
               consi += counter['cons']
               print 'Inconsistent'
            print '--------------------------------------'

            edges += counter['arcs']
            cons += counter['cons']
            counter['arcs'] = 0
            counter['cons'] = 0
            ConMatrix = None
          
   print '======================================'
   print 'Total time:', total, 'sec'
   print 'Avg time per CSP:', total/(1.0*meter), 'sec'
   if meter-meteri:
      print 'Avg time per path consistent CSP:', totalc/(1.0*(meter-meteri)), 'sec' 
   if meteri:
      print 'Avg time per inconsistent CSP:', totali/(1.0*meteri), 'sec'
   print 'Avg # of arcs processed per CSP:', edges/(1.0*meter)
   if meter-meteri:
      print 'Avg # of arcs processed per path consistent CSP:', edgesc/(1.0*(meter-meteri))
   if meteri:
      print 'Avg # of arcs processed per inconsistent CSP:', edgesi/(1.0*meteri)
   print 'Avg # of constraint checks per CSP:', cons/(1.0*meter)
   if meter-meteri:
      print 'Avg # of constraint checks per path consistent CSP:', consc/(1.0*(meter-meteri)) 
   if meteri:
      print 'Avg # of constraint checks per inconsistent CSP:', consi/(1.0*meteri)
   if meter-meteri:
      print '# of path consistent CSPs:', meter-meteri
   print '# of inconsistent CSPs:', meteri
   print '======================================'

if __name__ == '__main__':
   sys.exit(main())

